import java.text.SimpleDateFormat

import com.atlassian.jira.component.ComponentAccessor
import com.google.common.collect.LinkedListMultimap
import groovy.json.JsonSlurper
import org.apache.commons.codec.binary.Base64
import org.apache.commons.io.IOUtils
import org.apache.log4j.Logger

xray = new XrayTool(baseUrl: requestContext.canonicalBaseUrl)

/**
 * Velocity tool to get Xray test runs data from Xray REST API.
 */
class XrayTool {
	def log = Logger.getLogger(this.getClass())

	def baseUrl
	def customFieldManager = ComponentAccessor.customFieldManager
	def userManager = ComponentAccessor.userManager

	/* Xray REST API request parameters. */
	def restUserName = "admin"
	def restPassword = "admin"

	/**
	 * Some Xray REST end-points return dates in this hardwired format.
	 */
	def static xrayDateFormat = new SimpleDateFormat("yyyy-MM-dd'T'hh:mm:ssXXX", ComponentAccessor.jiraAuthenticationContext.locale)
	def jsonSlurper = new JsonSlurper()

	/**
	 * Returns the Xray test runs for the passed issue (as Groovy objects slurped from <code>https://confluence.xpand-it.com/display/public/XRAY/Tests+-+REST#Tests-REST-ExportingTestRuns<code>)
	 */
	def getTestRuns(def issue) {
		if(!isTest(issue)) {
			return []
		}

		def issueKey = issue.key
		log.debug("Downloading Xray test runs for issue #${issueKey}")
		def response = callRestApi("${baseUrl}/rest/raven/1.0/api/test/${issueKey}/testruns")
		if (response instanceof Throwable) {
			def apiNotFoundMessage = (response instanceof FileNotFoundException) ? " (do you have Xray version 2.0.0 or newer installed?)" : ""
			def failureMessage = "Failed to download Xray test runs${apiNotFoundMessage}: ${response.message}"
			return [["testExecKey": failureMessage]]
		} else {
			log.debug("${response.size()} Xray test runs found for issue #${issueKey}")
			return response
		}
	}

	def callRestApi(def url) {
		def connection
		def inputStream

		try {
			log.debug("Xray REST API URL: ${url}")
			connection = new URL(url).openConnection()
			def authValue = "${restUserName}:${restPassword}"
			connection.setRequestProperty("Authorization", "Basic " + Base64.encodeBase64String(authValue.getBytes("ISO-8859-1")))
			inputStream = connection.getInputStream()
			def json = IOUtils.toString(inputStream, "UTF-8")
			log.trace("Xray JSON for ${url}:\n${json}")

			def result = jsonSlurper.parseText(json)

			return result
		} catch (Throwable ex) {
			log.error("Failed to download Xray data", ex)
			return ex
		} finally {
			IOUtils.closeQuietly(inputStream)
		}
	}

  	def getTestSteps(def testRun) {
  		testRun?.steps ?: []
  	}

	def getKeys(def list) {
        list ? list.collect { it.key }.join(", ") : null
	}

	def getFileNames(def list) {
        list ? list.collect { it.fileName }.join(", ") : null
	}

	def getNames(def list) {
        list ? list.collect { it.name }.join(", ") : null
	}

	def getItems(def list) {
        list ? list.collect { it }.join(", ") : null
	}

	def getDate(def dateString) {
		dateString ? xrayDateFormat.parse(dateString) : null
	}

	def getDisplayName(def userNameOrKey) {
		if (userNameOrKey) {
			// expect user name, but fall back to user key as Xray may return that (unexpectedly) for certain versions and certain fields
			return userManager.getUserByName(userNameOrKey)?.displayName ?: userManager.getUserByKey(userNameOrKey)?.displayName
		} else {
			return null
		}
	}

	/**
	 * Returns whether the issue is an Xray test.
	 */
	private isTest(def issue) {
		// check whether the issue has a specific custom field type
		customFieldManager.getCustomFieldObjects(issue).find { it.customFieldType.key == "com.xpandit.plugins.xray:manual-test-steps-custom-field" }
	}

	static def formatDuration(def milliSeconds) {
		(milliSeconds != null) ? (milliSeconds / 1000.0 / 60.0 / 60.0) : null
	}

	/**
	 * Returns a "repeating" list of entries where the issue is the (repeating) key and its test runs is the value.
	 * Example: (i1 -> tr11, i1 -> tr12, i1 -> tr13, i2 -> tr21, i3 -> tr31, ...)
	 */
	def getRepeatingTestRuns(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each {
			it -> result.putAll(it, getTestRuns(it))
		}

		return result.entries()
	}
}
