import com.atlassian.jira.component.ComponentAccessor
import com.google.common.collect.LinkedListMultimap
import groovy.json.JsonSlurper
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import org.apache.commons.codec.binary.Base64
import org.apache.commons.io.IOUtils
import org.apache.commons.lang.StringEscapeUtils
import org.apache.log4j.Logger

zephyr = new ZephyrTool(issues: issues, baseUrl: requestContext.canonicalBaseUrl)

/**
 * Velocity tool to get Zephyr test data from ZAPI (Zephyr REST API).
 * Note: you must install the ZAPI app (https://marketplace.atlassian.com/plugins/com.thed.zephyr.zapi/server/overview) for this.
 *
 * @see https://zephyrdocs.atlassian.net/wiki/spaces/DEVELOPER/pages/33095703/ZAPI+Server
 */
class ZephyrTool {
	def log = Logger.getLogger(this.getClass())

	def issues
	def baseUrl
	def customFieldManager = ComponentAccessor.customFieldManager

	/* Zephyr REST API request parameters. */
	def restUserName = "admin"
	def restPassword = "admin"

	/**
	 * Stores the test executions for all issues after being pre-loaded with a single ZAPI call.
	 */
	def testExecutions

	/**
	 * Failure message that is set if there is an exception thrown while communicating with the Zephyr REST API, or <code>null</code> if there was no error.
	 */
	def failureMessage

	def jsonSlurper = new JsonSlurper()
	def xmlParser = new XmlParser()

	/**
	 * Some Zephyr REST end-points return dates in this hardwired format.
	 */
	def static zephyrDateFormat = new SimpleDateFormat("dd/MMM/yy", ComponentAccessor.jiraAuthenticationContext.locale)

	def static defectKeyComparator = { a, b ->
		def (aProjectKey, aNumber) = a.split("-")
		def (bProjectKey, bNumber) = b.split("-")
		(aProjectKey == bProjectKey) ? aNumber.toInteger() <=> bNumber.toInteger() : aProjectKey <=> bProjectKey
	}

	/**
	 * Returns the Zephyr test steps for the passed issue (as Groovy objects slurped from <code>http://docs.getzephyr.apiary.io/#reference/teststepresource/get-update-delete-teststep/get-teststep<code>)
	 */
	def getTestSteps(def issue) {
		if(!isTest(issue)) {
			return []
		}
		if(failureMessage) {
			return [["orderId": "", "step": "", "data": failureMessage, "result": ""]]
		}

		def issueId = issue.id
		log.debug("Downloading Zephyr test steps for issue #${issueId}")
		def response = callZapi("${baseUrl}/rest/zapi/latest/teststep/${issueId}")
		if (response instanceof Throwable) {
			failureMessage = "Failed to download Zephyr test steps (have you enabled the Zephyr API?): ${response.message}"
			return [["orderId": "", "step": "", "data": failureMessage, "result": ""]]
		} else {
			// handle JSON structure change ("response.stepBeanCollection" was introduced in Zephyr 4.0.0)
			if(response instanceof Map) {
				response = response.stepBeanCollection
			}
			log.debug("${response.size()} Zephyr test steps found for issue #${issueId}")
			return response
		}
	}

	/**
	 * Preloads and returns all Zephyr test executions, or returns the exception object if something went wrong.
	 */
	def preloadTestExecutions(def withStepResults) {
		def connection
		def inputStream

		def issueKeys = issues.findAll{ isTest(it) }.collect { it.key }
		def expand = (withStepResults) ? "teststeps" : ""
		def requestBody = """
			{
				"exportType": "xml",
				"maxAllowedResult": "true",
				"expand": "${expand}",
				"startIndex": "0",
				"zqlQuery": "issue in (${issueKeys.join(',')})",
				"isUI": false
			}
			""" // DO NOT change "maxAllowedResult" from string or "isUi" from boolean
		def response = callZapi("${baseUrl}/rest/zapi/latest/execution/export", "POST", requestBody)
		if(response instanceof Throwable) {
			failureMessage = "Failed to download Zephyr test executions (have you enabled the Zephyr API?): ${response.message}"
			return response
		}

		// return the forced result (empty list) if no test executions are found
		if((response instanceof Map) && response.containsKey("forcedResult")) {
			return response.forcedResult
		}

		try {
			connection = new URL(response.url).openConnection()
			def authValue = "${restUserName}:${restPassword}"
			connection.setRequestProperty("Authorization", "Basic " + Base64.encodeBase64String(authValue.getBytes("ISO-8859-1")))
			inputStream = connection.getInputStream()
			def xml = IOUtils.toString(inputStream, "UTF-8")
			log.trace("Original XML:\n${xml}")
			xml = fixXml(xml)
			log.trace("Fixed XML:\n${xml}")
			def testExecutions = xmlParser.parseText(xml)
			log.debug("${testExecutions.execution.size()} Zephyr test executions found")
			return testExecutions
		} catch (Throwable ex) {
			log.error("Failed to download Zephyr XML data", ex)
			failureMessage = "Failed to download Zephyr test executions (have you enabled the Zephyr API?): ${ex.message}"
			return ex
		} finally {
			IOUtils.closeQuietly(inputStream)
		}
	}

	/**
	 * Returns the Zephyr test executions (optionally with step results) for the passed issue.
	 * Impl. note: this processes the XML returned by Zephyr by addressing the XML element by their index, because the tag names depend on the user-selected locale (ZAPI design decision).
	 */
	def getTestExecutions(def issue, def withStepResults) {
		def result = []

		if (!isTest(issue)) {
			return result
		}
		if (!testExecutions) {
			testExecutions = preloadTestExecutions(withStepResults)
		}
		if (!(testExecutions instanceof Throwable)) {
			testExecutions.findAll { it.children()[2].text() == issue.key }.each {
				def stepResults = []
				if (withStepResults) {
					def testStepsNode = it.children().find { it.name() == "teststeps"} // index of test steps node was changed in Zephyr 4.0.0
					testStepsNode.children().each {
						stepResults.add([
							"order"         : it.children()[1].text().toInteger(),
							"step"          : it.children()[2].text(),
							"data"          : it.children()[3].text(),
							"expectedResult": it.children()[4].text(),
							"status"        : it.children()[5].text(),
							"defects"       : "", // not returned by this REST endpoint
							"comment"       : it.children()[6].text()
						])
					}
				}

				def defects = it.children()[14].text().split("\\|")
				result.add([
					"id"              : it.children()[0].text().toInteger(),
					"cycle"           : it.children()[1].text(),
					"version"         : it.children()[6].children()[0].text(),
					"status"          : it.children()[9].text(),
					"assignee"        : it.children()[11].text(),
					"executedBy"      : it.children()[12].text(),
					"executedOn"      : parseZephyrDate(it.children()[13].text()),
					"executionDefects": sortDefectKeys(defects.getAt(0)),
					"stepDefects"     : (defects.length > 1) ? sortDefectKeys(defects.getAt(1)) : "",
					"comment"         : xmlToDelimitedString(StringEscapeUtils.unescapeXml(it.children()[10].text()), " "),
					"stepResults"     : stepResults
				])
			}
		} else {
			result.add([
				"id"              : "",
				"key"             : "",
				"version"         : "",
				"cycle"           : failureMessage,
				"assignee"        : "",
				"status"          : "",
				"executionDefects": "",
				"stepDefects"     : "",
				"executedBy"      : "",
				"executedOn"      : "",
				"comment"         : "",
				"stepResults"     : [["order": "", "step": "", "data": "", "expectedResult": "", "status": "", "defects": "", "comment": ""]]
			])
		}
		log.trace("Result: ${result}")

		return result
	}

	def callZapi(def url, def mode = "GET", def requestBody = "") {
		def connection
		def inputStream
		def errorStream

		try {
			log.debug("Zephyr URL: ${url}")
			connection = new URL(url).openConnection()
			def authValue = "${restUserName}:${restPassword}"
			connection.setRequestProperty("Authorization", "Basic " + Base64.encodeBase64String(authValue.getBytes("ISO-8859-1")))
			if(mode == "POST") {
				connection.setRequestMethod("POST")
				connection.setDoOutput(true)
				connection.setRequestProperty("Content-Type", "application/json")
				connection.getOutputStream().write(requestBody.getBytes("UTF-8"))
			}

			// replace the 406 status code (returned by ZAPI somewhat unexpectedly) with an empty list when no test executions found at all
			if(connection.getResponseCode() == 406) {
				errorStream = connection.getErrorStream()
				def json = IOUtils.toString(errorStream, "UTF-8")
				log.trace("Zephyr JSON for ${url}:\n${json}")
				def result = jsonSlurper.parseText(json)
				if(result.warning == "No matching record found.") {
					return ["forcedResult" : []]
				}
			}

			inputStream = connection.getInputStream()
			def json = IOUtils.toString(inputStream, "UTF-8")
			log.trace("Zephyr JSON for ${url}:\n${json}")

			def result = jsonSlurper.parseText(json)

			return result
		} catch (Throwable ex) {
			log.error("Failed to download Zephyr data", ex)
			return ex
		} finally {
			IOUtils.closeQuietly(inputStream)
			IOUtils.closeQuietly(errorStream)
		}
	}

	/**
	 * Fixes problems in XML produced by Zephyr, or does nothing if the syntax is correct:
	 * - Invalid CDATA syntax
	 * 		Invalid format: "![CDATA[Sample text.]]"
	 * 		Valid format: "<![CDATA[Sample text.]]>"
	 *
	 * - Invalid tag names (e.g. "Foo bar" or even "Foo (bar)"
	 */
	private static fixXml(def xml) {
		// fix CDATA sections
		xml = (xml.contains(">![CDATA[")) ? xml.replaceAll(/!\[CDATA\[/, "<![CDATA[").replaceAll(/\]\]<\//, "]]></") : xml

		// fix tag names with invalid characters
		xml.findAll(~/<.*?>/).unique().each {
			if(!it.contains("CDATA") && !it.contains("xml version")) {
				def pattern = Pattern.compile("[^\\w<>\\[\\]/]", Pattern.UNICODE_CHARACTER_CLASS)
				def correctTag = it.replaceAll(pattern, "")
				pattern = ~"\\Q${it}\\E"
				xml = xml.replaceAll(pattern, correctTag)
			}
		}

		return xml
	}

	/**
	 * Returns whether the issue is a Zephyr test.
	 */
	private isTest(def issue) {
		// check whether the issue has a specific custom field type
		customFieldManager.getCustomFieldObjects(issue).find { it.customFieldType.key == "com.thed.zephyr.je:zephyr-je-customfield-teststep" }
	}

	/**
	 * Parses (unsorted) defect keys from a CSV string, sorts them and returns them as CSV string.
	 */
	private static sortDefectKeys(def defectsString) {
		defectsString.split(",").collect{ it.trim() }.sort(defectKeyComparator).join(", ")
	}

	/**
	 * Parses a Zephyr date string to Excel time millis.
	 */
	private parseZephyrDate(def dateString) {
		def excelTime
		try {
			excelTime = (dateString) ? (25569 + zephyrDateFormat.parse(dateString).getTime() / (24 * 60 * 60 * 1000)).doubleValue().round() : ""
		} catch(Throwable ex) {
			log.error("Failed to parse Zephyr date", ex)
		}
		(excelTime) ?: dateString
	}

	/**
	 * Returns the body text fragments from the passed XML joined with the delimiter.
	 * The XML is not required to be well-formatted and can even be plain text.
	 */
	private static xmlToDelimitedString(def xml, def delimiter) {
		xml.split(/<.*?>/).collect{ it.trim() }.findAll{ !it.allWhitespace }.join(delimiter)
	}

	/**
	 * Returns a "repeating" list of entries where the issue (test) is the (repeating) key and its execution is the value.
	 * Example: (i1 -> e11, i1 -> e12, i1 -> e13, i2 -> e21, i3 -> e31, ...)
	 * Used by zephyr-report.xlsx template
	 */
	def getRepeatingExecutions(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each {
			it -> result.putAll(it, getTestExecutions(it, false))
		}

		return result.entries()
	}
}
